local addon = LibStub( "AceAddon-3.0" ):NewAddon( "SimpleActionSets", "AceEvent-3.0" )
SimpleActionSets = addon
local L = LibStub( "AceLocale-3.0" ):GetLocale( "SimpleActionSets" )
--TODO localization support

local LSM = LibStub("LibSharedMedia-3.0")

local db, temp
local debug = false
local version, build, date, tocversion = GetBuildInfo()

-- Setup slash command to access menu
SLASH_SAS1 = '/sas';
function SlashCmdList.SAS(msg, editbox)
	local command, set = msg:match("^(%S*)%s*(.-)$");
	SASDebug(command.." | "..set);
	
	if (command == L["SAS_TEXT_CONSOLE_LOAD"] and set:len() > 0) then
		if ( set and addon:SetExists( set ) ) then
			addon:SwapSet( set );
		else
			SASPrint( string.gsub(L["SAS_TEXT_CONSOLE_NOVALID"], "%%s", set) );
		end
	elseif (command == L["SAS_TEXT_CONSOLE_SAVE"] and set:len() > 0) then 
		if (addon:SetExists( set ) ) then
			-- do we need to delete it first?
			addon:SaveSet( set );
		else
			addon:SaveSet( set );
		end
	elseif (command == L["SAS_TEXT_CONSOLE_LIST"] and set:len() == 0) then 
		for k,v in pairs(db.sets) do SASPrint(k) end 
	elseif (command == L["SAS_TEXT_CONSOLE_DELETE"] and set:len() > 0) then 
		if (addon:SetExists( set) ) then
			addon:Delete( set );
		else
			SASPrint("No such set");
		end
	--TODO enable showing interface
	--else if ((command == L["SAS_TEXT_CONSOLE_SHOW"] and set:len() == 0) || command == nil) then
	else
		SASPrint(L["SAS_TEXT_CONSOLE_INVALID"]..command.." "..set.."\n"..L["SAS_TEXT_CONSOLE_HELP"]);
	end
end

addon.defaults = { 
	profile = {
		noUIWarnings = false,
		emptyBars = false,
		noEmptyButtons = false,
		hideFakeItemTooltips = false,
		autoRestore = false,
		noAutoRestoreWarning = false,
		noRemove = false,
		miniMapButton = {
			hide = false,
			lock = false,
			position = 270,
			offset = 0
		},
		currentSet = "",
		sets = {},
	}
}

function addon:OnInitialize()
    -- DB
	self.db = LibStub( "AceDB-3.0" ):New( "SAS_Saved", self.defaults )
    db = addon.db.profile
end
function addon:OnEnable()
	SASPrint("Thank for for using "..L["SAS_TITLE"]);
end
function addon:OnDisable()
	
end

-------------------
-- Load functions
--------------------
function addon:SwapSet( set )
	local actions;
	if ( set ) then
		SASDebug(" SAS_SwapSet - about to do copy table on "..set );
		actions =addon:CopyTable( db.sets[set] );
	elseif ( SASMain:IsVisible() ) then
		SASDebug(" SAS_SwapSet - about to do copy table on SAS_Temp" );
		actions =addon:CopyTable( SAS_Temp );
	else
		return;
	end

	if ( not actions ) then
		SASDebug("|wNo actions to change to.");
		return;
	end

	if ( not inCombat ) then
		addon:DoSwap(actions, set);
	else
		SASPrint(L["SAS_TEXT_WARNING_COMBAT"]);
	end
end

function addon:DoSwap(actions, set, player)
	-- Swaps a saved set to the player's real action bars.
	SAS_SwappingSet = true;
	if ( not player or player == PlrName ) then
		addon:SetCurrentSet( set );
	end
	ClearCursor();
	for i = 0, 9 do
		if (addon:BarEnabled(i,actions) and not ( not addon:BarHasActions(i, actions) and not db.emptyBars ) ) then
			SASDebug("Bar "..(i+1).." is enabled");
			for j = 1, 12 do
				local id = j + i*12;
				local msg = "Action "..id;
				if ( actions[i][j]  ) then
					if ( actions[i][j] == 0 ) then
						msg = msg.." forced cleared.";
						addon:ClearSlot(id);
					elseif ( actions[i][j] == 1 ) then
						msg = msg.." forced no change.";
					else
						local ename, etexture, erank, elink, emacro =addon:GetActionInfo(id);
						local name, texture, rank, link, macro =addon:ParseActionInfo( actions[i][j] ); --actions[i][j][1], actions[i][j][2], actions[i][j][3], actions[i][j][4], actions[i][j][5];
						if ( (name == ename and (( rank == erank ) or ( macro and emacro ) or ( link and elink ))) ) then
							msg = msg.." is the same as the one trying to be swapped.";
						else						
							if ( macro ) then							
								msg = msg.." is a macro.";
								local macroID =addon:FindMacro( name, texture, macro )
								if ( macroID ) then
									PickupMacro( macroID );
									PlaceAction(id);
								else
									msg = msg.." |wCouldn't find macro."
									addon:ClearSlot(id);
								end
							elseif ( link ) then							
								msg = msg.." is an item.";
								if ( not addon:PlaceItem( id, link, name, set ) ) then
									msg = msg.." |wCouldn't find item."
									addon:MissingItem( id, {name, texture, link, set} );
								end
							elseif ( rank == "MOUNT" or rank == "CRITTER" ) then
								msg = msg.." is a companion.";
								local companionId = addon:FindCompanion( name, rank );
								if ( companionId ) then
									PickupCompanion( rank, companionId );
									PlaceAction(id);
								else
									msg = msg.." |wCouldn't find companion."
									addon:ClearSlot(id);
								end
							elseif ( name ) then
								msg = msg.." is a spell.";
								local spellNum, highest =addon:FindSpell( name, rank );
								if ( spellNum ) then
									PickupSpellBookItem( spellNum, BOOKTYPE_SPELL );
									PlaceAction(id);
								elseif ( highest ) then
									local spellName, spellRank =  GetSpellBookItemName( highest, BOOKTYPE_SPELL );
									if ( rank ~= spellRank ) then
										msg = msg.." Can't find rank, using highest found instead.";
										PickupSpellBookItem( highest, BOOKTYPE_SPELL );
										PlaceAction(id);
									else
										msg = msg.." Action is highest rank found, won't swap.";
									end
								else
									msg = msg.." |wCoulnd't find spell."
									addon:ClearSlot(id);
								end
							else
								msg = msg.." is has no name, link, macro?";
							end
						end

						ClearCursor();
					end
				elseif ( not db.noEmptyButtons ) then
					msg = msg.." is empty.";
					addon:ClearSlot(id);
				end
				SASDebug(msg);
			end
		elseif ( not addon:BarEnabled(i,actions) ) then
			SASDebug("Bar "..(i+1).." not enabled, skipping");
		else
			SASDebug("Bar "..(i+1).." has no actions, skipping");
		end
	end
	ClearCursor();

	SAS_SwappingSet = nil;
end


-----------------------
-- Save function
---------------------

function addon:SaveSet(set)
	-- Save a set
	if ( not set ) then
		if (addon:GetCurrentSet() ) then
			set =addon:GetCurrentSet();
		else
			return;
		end
	else
		addon:SetCurrentSet(set);
	end
	local actions;
	if ( SAS_Temp ) then
		actions =addon:CopyTable(SAS_Temp);
	else
		actions =addon:IterateActions();
	end
	SASPrint(L["SAS_TEXT_SAVING"]..set);
	db.sets[set] =addon:CopyTable(actions);
end

function SASActions_SaveNew( set )
	SASSaveMenuNameEB:SetText("");
	ShowUIPanel(SASSaveMenu);
	if ( set ) then
		SASSaveMenuNameEB:SetText( set );
		SASSaveMenu_Save( set );
	end
end

function addon:IterateActions()
	-- Iterate over all current actions
	local actionlist = {};
	for i=0, 9 do
		actionlist[i] = {};
		for j=1, 12 do
			actionlist[i][j] =addon:BuildActionInfo( addon:GetActionInfo( j+(i*12) ) ); --action[1], action[2], action[3], action[4], action[5] );
		end
	end

	return actionlist;
end

function addon:BuildActionInfo(...) -- name, texture, rank, link, macro, ?
	local temp, string, i = "";
	for i=1, select("#",...) do
		local val = select(i,...);
		if ( val ) then
			string = (string or "")..temp..(val or "").."";
			temp = "";
		else
			temp = temp .. "";
		end
	end
	return string;
end

function addon:GetActionInfo( id )
	local actionType, typeId, subType = GetActionInfo(id);
		
	if ( actionType == "spell" and typeId > 0 ) then
		local name, rank, icon =  GetSpellInfo(typeId);
		return name, (icon), rank;
	elseif ( actionType == "item" ) then
		local texture = (GetActionTexture(id));
		return (GetItemInfo(typeId)), texture, nil, typeId;
	elseif ( actionType == "macro" ) then
		local name, texture = GetMacroInfo(typeId);
		if ( not texture ) then
			texture = (GetActionTexture(id));
		else
			texture = (texture);
		end
		return name, texture, nil, nil, typeId;
	elseif ( actionType == "companion" ) then
		local npcid, name, spellid, texture = GetCompanionInfo( subType, typeId );
		--TODO figure out why getcompanion info is broken
		return typeId, nil, subType; -- treat companions as spells for now
	end
end

-------------------------------
-- Utility Functions
-------------------------------
function addon:ParseActionInfo( action, ... )
	if not action then return; end
	if type(action)~="string" then return action; end
	local a = {strsplit("",action)};
	for k, v in pairs(a) do if v == "" then a[k] = nil; end end
	if ( select("#",...) > 0 ) then
		local b = {};
		for i=1, select("#",...) do
			tinsert( b, a[select(i,...)] );
		end
		return unpack(b);
	else
		if a[3] ~= "MOUNT" and a[3] ~= "CRITTER" then
			a[3] = tonumber(a[3])
		end
		return a[1], a[2], a[3], a[4], tonumber(a[5]), a[6]; --name, texture, rank, link, macro, ?;
	end
end

function addon:BarEnabled( bar, set )
	if ( not set ) then set = SAS_Temp; end
	if ( set and set[bar] ) then
		return not set[bar][0];
	end
end

function addon:FindMacro( name, texture, macro )
	-- Check saved macro id
	local macroName, macroTexture, bestguess;
	if ( macro ) then
		macroName, macroTexture = GetMacroInfo(macro);
		if ( macroName == name ) then
			if ( texture and macroTexture == texture ) then
				return macro;
			end
			bestguess = macro;
		end
	end

	-- If no direct match, iterate over macros and return matching index
	local numAccountMacros, numCharacterMacros = GetNumMacros();
	numCharacterMacros = numCharacterMacros + 36;
	for i=1, 36+18 do
		if ( i > numAccountMacros and i < 37 ) then
			-- no more global macros, skip to local
			i = 37;
		end
		if ( i > numCharacterMacros ) then
			-- no more macros
			break;
		end
		macroName, macroTexture = GetMacroInfo(i)
		if ( macroName and macroName == name ) then
			if ( texture and macroTexture == texture ) then
				return i;
			end
			bestguess = i;
		end
	end

	return bestguess;
end

function addon:BarHasActions(bar, set)
	if ( not set ) then
		set = SAS_Temp;
	end
	if ( bar and set[bar] ) then
		for i=1, 12 do
			if ( set[bar][i] ) then
				return 1;
			end
		end
	end
end

function addon:SetCurrentSet( set )
	SASDebug("Changing current set to " .. (set or "nil"));
	db.currentSet = set;
end

function SASActions_ClearTemp()
	-- Clear the temp set
	for i=0, 9 do
		SAS_Temp[i] = {};
	end
	SASActions_Display();
end

function addon:SetExists( set )
	-- Check to see if the set already exists
	if db.sets[set] then
		SASDebug(set.." exists");
		return true;
	end
	SASDebug(set.." does not exist");
end

function addon:CopyTable(copyTable)
	-- properly copies a table instead of referencing the same table, thanks Sallust.
	if ( not copyTable ) then return; end
	local returnTable = {};
	for k, v in pairs(copyTable) do
		if type(v) == "table" then
			returnTable[k] =addon:CopyTable(v);
		else
			returnTable[k] = v;
		end
	end
	return returnTable;
end

function addon:ClearSlot(id)
	-- Clear an action slot
	if ( HasAction(id) and not db.noEmptyButtons ) then
		PickupAction(id);
		ClearCursor();
	end
end

local function actionComp( a, b )
	if a and b then
		local A = {addon:ParseActionInfo( a ) };
		local B = {addon:ParseActionInfo( b ) };
		for k, v in pairs(A) do
			if ( k ~= 4 and (not B[k] or v ~= B[k]) ) then
				SASDebug( k.." "..v.." does not == "..B[k] );
				return 1;
			end
		end
	else
		if ( a ) then
			SASDebug( "no b" );
		elseif ( b ) then
			SASDebug( "no a" );
		else
			SASDebug( "no a or b" );
		end
		return 1;
	end
end

function addon:FindSpell( spell, rank )
	if ( not spell ) then return; end
	-- Iterate over spells the player has and return location
	local i, name, texture, highest = 1;
	local spellName, spellRank =  GetSpellBookItemName( i, BOOKTYPE_SPELL );
	while spellName do
		if ( spellName == spell ) then
			highest = i;
			if ( rank ) then
				if ( tRank(spellRank) == tonumber(rank) ) then
					return i;
				end
			else
				return i;
			end
		end
		i = i+1;
		spellName, spellRank =  GetSpellBookItemName( i, BOOKTYPE_SPELL );
	end
	return nil, highest;
end

function addon:CompareSet( set1, set2, strict )
	-- Compare two sets to see if they're similar
	if ( not set1 or not set2 ) then
		if ( not set1 ) then
			SASDebug( "no set1" );
		elseif ( not set2 ) then
			SASDebug( "no set2" );
		else
			SASDebug( "no set1 or set2" );
		end
		return 1;
	end

	for i=0, 9 do
		if ( set1[i] ) then
			if ( strict == 2 ) then
				if ( set1[i][0] ~= set2[i][0] ) then
					return 1;
				end
			end
			for j=1, 12 do
				if ( set1[i][j] ) then
					if ( actionComp( set1[i][j], set2[i][j] ) ) then
						SASDebug( "returning at "..i.." "..j );
						return 1;
					end
				elseif ( strict and set2[i][j] ) then
					SASDebug( "strict returning at "..i.." "..j.." no set1 but set2" );
					return 1;
				end
			end
		elseif ( strict and set2[i] ) then
			SASDebug( "strict returning at "..i.." no set1 but set2" );
			return 1;
		end
	end
end

function addon:Delete(set)
	-- Delete a set
	SASDebug(L["SAS_TEXT_DELETING"]..set);
	db.sets[set] = nil;
end

function addon:PlaceItem( id, link, name, set )
	-- Place an item, update set link id if found
	local itemLink, bag, slot, inv;
	if ( link == "?" or link == "1" ) then
		itemLink, bag, slot, inv = addon:CheckItem( name, id, set );
	else
		itemLink, bag, slot, inv = addon:FindItem( link );
	end
	if ( bag ) then
		PickupContainerItem( bag, slot );
		PlaceAction(id);
		return 1;
	elseif ( inv ) then
		PickupInventoryItem( inv );
		PlaceAction(id);
		return 1;
	end
end

function addon:CheckItem( name, id, set )
	if ( name ) then
		itemLink, bag, slot, inv = addon:FindItem( name );
		itemLink = addon:FindLink(itemLink)
		if ( not itemLink ) then
			itemLink = addon:CheckItemDBs( name );
		end
		if ( itemLink and itemLink ~= "?" and itemLink ~= "1" ) then
			local slot = math.mod( id, 12 );
			local bar = (id-slot) / 12;
			if ( db.sets[set] ) then
				local action = db.sets[set][bar][slot];
				local iname, ilink = addon:ParseActionInfo(action, 1, 4);
				if ( iname ) then
					if ( iname == name and ilink and ilink == "?" ) then
						SASDebug( "Updating item in action #"..id.." in set "..set.." with itemLink." );
						db.sets[set][bar][slot] = addon:IncActionInfo(action, itemLink, 4);
					else
						SASDebug( "|wAttempted to update itemLink in action #"..id.." but names are wrong?" );
					end
				end
			end
		end
	end
	return itemLink, bag, slot, inv;
end

function addon:FindItem( item )
	-- Iterate over items the player has and return it's link and location --
	if ( not item ) then return; end

	item = tostring(item);

	-- Iterate over bags
	for i=0, 4 do
		local bagSlots = GetContainerNumSlots(i);
		if ( bagSlots ) then
			for j=1, bagSlots do
				SASDebug("Doing GetContainerItemLink() on "..i.." "..j);
				local itemLink = GetContainerItemLink(i,j);
				if ( itemLink ) then
					if ( item == addon:FindLink( itemLink ) or item == addon:FindName( itemLink ) ) then
						return itemLink, i, j;
					end
				end
			end
		end
	end

	-- Iterate over paper doll
	for i=0, 23 do
		SASDebug("Doing GetInventoryItemLink() on "..i);
		local itemLink = GetInventoryItemLink("player",i);
		if ( itemLink ) then
			if ( itemLink ) then
				if ( item == addon:FindLink( itemLink ) or item == addon:FindName( itemLink ) ) then
					return itemLink, nil, nil, i;
				end
			end
		end
	end
end

function addon:FindLink( item )
	-- Find an item's link number from it's item link
	return string.match( item, "item:(%d+):" );
end

function addon:ItemName( item )
	-- Return an item's name
	local itemName = addon:FindItem( item );
	if ( itemName ) then
		return addon:FindName( itemName );
	end
end
function addon:FindName( item )
	-- Find an item's name from it's item link
	if ( item ) then
		for name in string.gmatch( item, "%[(.+)%]") do
			return name;
		end
	end
end

function addon:MissingItem( id, itemInfo )
	-- Add to list of items to place on the bar when available.
	if ( id and itemInfo ) then
		if ( not db["MissingItems"] ) then
			db["MissingItems"] = {};
		end
		local missing = db["MissingItems"]
		if ( not missing[id] or addon:ParseActionInfo(missing[id], 3) ~= itemInfo[3] ) then
			db["MissingItems"][id] = addon:BuildActionInfo(itemInfo[1], itemInfo[2], nil, itemInfo[3], nil, nil, itemInfo[4]);
			if ( HasAction( id ) ) then
				PickupAction( id );
				ClearCursor();
			end
			SASDebug( itemInfo[1] );
		end
	end
end

--TODO change back to using companion name in place of companion id
function addon:FindCompanion( match, companionType )
	local i, companionId, companionName, spellId;
	for i=1, GetNumCompanions(companionType) do
		companionId, companionName, spellId = GetCompanionInfo(companionType, i);
		if match == tostring(spellId) then
			-- need to return the index
			return i, spellId;
		end
	end
end



---------------------------------
-- Output mapping functions 
---------------------------------

function SASPrint(msg)
	-- Basic print function
	msg = tostring(msg)
	msg = string.gsub( msg, "%|w", "|cffff0000" );
	msg = string.gsub( msg, "%|i", "|cffe6e6fA" );
	DEFAULT_CHAT_FRAME:AddMessage( "<|cff40E0D0SAS|r> "..msg, 0.012, 0.658, 0.62 );
	--local cols = "|cff03A98E";
end

function SASDebug(msg)
	-- Print when debug is turned on
	if ( debug ) then SASPrint("<|wDEBUG|r> "..tostring(msg)); end
end

